/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	Copyright 1998-2002 Anton Vinokurov <anton@netams.com>
***	Copyright 2002-2008 NeTAMS Development Team
***	This code is GPL v3
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: ds_libpcap.c 1628 2013-10-28 19:08:22Z anton $ */

#include "config.h"
#include <pcap.h>
extern u_int64_t flowid;
#ifndef NOPCAP

//obtained from fprobe
struct DLT dlt[] = {
#ifdef DLT_NULL
        {DLT_NULL, 0, 4, 4, (char*)"NULL"},
#endif
#ifdef DLT_EN10MB
        {DLT_EN10MB, 12, 14, 17, (char*)"EN10MB"},
#endif
#ifdef DLT_IEEE802
        {DLT_IEEE802, 14, 22, 17, (char*)"IEEE802"},
#endif
#ifdef DLT_ARCNET
        {DLT_ARCNET, 2 , 6, 6, (char*)"ARCNET"},
#endif
#ifdef DLT_SLIP
        {DLT_SLIP, -1, 16, 16, (char*)"SLIP"},
#endif
#ifdef DLT_PPP
        {DLT_PPP, 2, 4, 4,(char*)"PPP"},
#endif
#ifdef DLT_FDDI
        /*
        FIXME
        See gencode.c from libpcap
        */
        {DLT_FDDI, 13, 21, 16, (char*)"FDDI"},
#endif
#ifdef DLT_ATM_RFC1483
        {DLT_ATM_RFC1483, 0, 8, 3, (char*)"ATM_RFC1483"},
#endif
#ifdef DLT_RAW
        {DLT_RAW, -1, 0, 0, (char*)"RAW"},
#endif
#ifdef DLT_SLIP_BSDOS
        {DLT_SLIP_BSDOS, -1, 24, 24, (char*)"SLIP_BSDOS"},
#endif
#ifdef DLT_PPP_BSDOS
        {DLT_PPP_BSDOS, 5, 24, 24, (char*)"PPP_BSDOS"},
#endif
#ifdef DLT_ATM_CLIP
        {DLT_ATM_CLIP, 0, 8, 3, (char*)"ATM_CLIP"},
#endif
#ifdef DLT_PPP_SERIAL
        {DLT_PPP_SERIAL, 2, 4, 4, (char*)"PPP_SERIAL"},
#endif
#ifdef DLT_PPP_ETHER
        {DLT_PPP_ETHER, 6, 8, 8, (char*)"PPP_ETHER"},
#endif
#ifdef DLT_C_HDLC
        {DLT_C_HDLC, 2, 4, 4, (char*)"C_HDLC"},
#endif
#ifdef DLT_IEEE802_11
        {DLT_IEEE802_11, 24, 32, 27, (char*)"IEEE802_11"},
#endif
#ifdef DLT_LOOP
        {DLT_LOOP, 0, 4, 4, (char*)"LOOP"},
#endif
#ifdef DLT_LINUX_SLL
        {DLT_LINUX_SLL, 14, 16, 16, (char*)"LINUX_SLL"},
#endif
#ifdef DLT_LTALK
        {DLT_LTALK, -1, 0, 0, (char*)"LTALK"},
#endif
#ifdef DLT_PRISM_HEADER
        {DLT_PRISM_HEADER, 144 + 24, 144 + 30, 144 + 27, (char*)"PRISM_HEADER"},
#endif
#ifdef DLT_IP_OVER_FC
        {DLT_IP_OVER_FC, 16, 24, 19, (char*)"IP_OVER_FC"},
#endif
#ifdef DLT_SUNATM
        {DLT_SUNATM, 4, 4 + 8, 4 + 3, (char*)"SUNATM"},
#endif
#ifdef DLT_ARCNET_LINUX
        {DLT_ARCNET_LINUX, 4, 8, 8, (char*)"ARCNET_LINUX"},
#endif
#ifdef DLT_ENC
        {DLT_ENC, 0, 12, 12, (char*)"ENC"},
#endif
#ifdef DLT_FRELAY
        {DLT_FRELAY, -1, 0, 0, (char*)"FRELAY"},
#endif
#ifdef DLT_IEEE802_11_RADIO
        {DLT_IEEE802_11_RADIO, 64 + 24, 64 + 32, 64 + 27, (char*)"IEEE802_11_RADIO"},
#endif
#ifdef DLT_PFLOG
        {DLT_PFLOG, 0, 28, 28, (char*)"PFLOG"},
#endif
#ifdef DLT_LINUX_IRDA
        {DLT_LINUX_IRDA, -1, -1, -1, (char*)"LINUX_IRDA"},
#endif
#ifdef DLT_APPLE_IP_OVER_IEEE1394
        {DLT_APPLE_IP_OVER_IEEE1394, 16, 18, 0, (char*)"APPLE_IP_OVER_IEEE1394"},
#endif
#ifdef DLT_IEEE802_11_RADIO_AVS
        {DLT_IEEE802_11_RADIO_AVS, 64 + 24, 64 + 32, 64 + 27, (char*)"IEEE802_11_RADIO_AVS"},
#endif
#ifdef DLT_PFSYNC
        {DLT_PFSYNC, -1, 4, 4, (char*)"PFSYNC"},
#endif
        {-1, -1, -1, -1, (char*)"UNKNOWN"}
};


void ds_libpcap_start(struct ds_pcap_info *pcap_info){
	u_char i;
	char errbuf[PCAP_ERRBUF_SIZE];

	while ((pcap_info->pcap = pcap_create(pcap_info->interface, errbuf))==NULL){
		n4debug("datasource @@ failed to create pcap interface [%s]: %s\n", pcap_info->interface, errbuf);
		sleep(10);
	}
	n4debug("datasource @@ interface '%s'\n", pcap_info->interface);

	if (pcap_info->pcap_bufsize>0) {
		i = pcap_set_buffer_size(pcap_info->pcap, pcap_info->pcap_bufsize*1024*1024);
		n4debug("datasource @@ setbufsize [%d Mb] %d\n", pcap_info->pcap_bufsize, i);
	}
	
	int snaplen=60;
	if (pcap_info->enable_url_check) snaplen=1514;
	if (pcap_info->enable_dpi) snaplen=65535;
	i = pcap_set_snaplen(pcap_info->pcap, snaplen);
	n4debug("datasource @@ setsnaplen [%d] %d\n", snaplen, i);

	pcap_set_promisc(pcap_info->pcap, 1);
	pcap_set_timeout(pcap_info->pcap, 1000);
	pcap_activate(pcap_info->pcap);
	        
//	while((pcap_info->pcap = pcap_open_live(pcap_info->interface, 0xffff, 1, 1000, errbuf))==NULL){
//		n4debug("datasource @@ failed to open pcap interface: %s\n", errbuf);
//		sleep(10);
//	}

	bpf_u_int32 netp, maskp;
	if(pcap_lookupnet(pcap_info->interface, &netp, &maskp, errbuf) == -1) {
		n4debug("datasource @@ lookupnet failed %s\n", errbuf);
	}

	pcap_info->pcap_datalink = pcap_datalink(pcap_info->pcap);
	for (i = 0;; i++)
		if (dlt[i].linktype == pcap_info->pcap_datalink || dlt[i].linktype == -1) {
			pcap_info->pcap_offset = dlt[i].offset_nl;
			pcap_info->link_type_idx = i;
			break;
		}
	n4debug("datasource @@ RUN libpcap %s: %s\n", pcap_info->interface, dlt[i].descr);


	if (pcap_info->filter) {
		if (-1 == pcap_compile(pcap_info->pcap, &pcap_info->pf, pcap_info->filter, 1, 0)) {
			n4debug("datasource @@ failed to compile pcap: %s\n", pcap_geterr(pcap_info->pcap));
			return;
		}
		if (-1 == pcap_setfilter(pcap_info->pcap, &pcap_info->pf)) {
			n4debug("datasource @@ failed to setfilter pcap: %s\n", pcap_geterr(pcap_info->pcap));
			return;
		}

#ifndef NETBSD
		pcap_freecode(&pcap_info->pf);
#endif
	} else
		n4debug("datasource @@ no rule for pcap, capturing all \n");

	pcap_info->fd=pcap_get_selectable_fd(pcap_info->pcap);
	pcap_info->fc = new flowcache(pcap_info);

#ifdef ENABLE_DPI
	init_dpi();
#endif

	n4debug("datasource @@ Libpcap packet processing for data-source initialized, fd=%d\n", pcap_info->fd);

}

void ds_pcap_callback(unsigned char *args, const struct pcap_pkthdr *hdr, const unsigned char *packet){
	struct ds_pcap_info *pcap_info = (struct ds_pcap_info *)args;

	u_short			ether_type;
	struct ip		*ip;
	struct pcap_ipv4_key	pcap_flow_key;

	if (pcap_info->is_registered==0) return;

	memset(&pcap_flow_key, '\0', sizeof(struct pcap_ipv4_key));

	switch (pcap_info->pcap_datalink) {
#ifdef DLT_EN10MB
		case DLT_EN10MB: {
			register const struct ether_header *ep;
			ep = (struct ether_header *)packet;
			ether_type = ntohs(ep->ether_type);
			packet += ETHER_HDR_LEN;
			break;
		}
#endif
#ifdef DLT_LINUX_SLL
		case DLT_LINUX_SLL: {
			if (hdr->len < SLL_HDR_LEN) return;     /* This should never happened */

			register const struct sll_header *sllp;
			sllp = (struct sll_header *)packet;

			ether_type = ntohs(sllp->sll_protocol);
			if (ether_type <= ETHERMTU) return; /* We got LLC layer. I don't know if IP is possible at this level */
            packet += SLL_HDR_LEN;
            break;
		}
#endif
		default:
        	ether_type = 0xFFFF;
            break;
	}

	if (ether_type != 0xFFFF) {
recurse:
		if (ether_type == ETHERTYPE_VLAN) {
			ether_type = ntohs(*(u_int16_t *) (packet + 2));
			pcap_flow_key.dot1x = *(u_int16_t *)packet;
			packet += 4;
			if (ether_type > ETHERMTU)
				goto recurse;

		}
		if (ether_type != ETHERTYPE_IP) return;

		ip = (struct ip *)packet;
	} else
		ip = (struct ip *)(packet + pcap_info->pcap_offset);

	IPv4GetKey(ip, (struct ipv4_key*)&pcap_flow_key);
	pcap_info->total_packets++;
	pcap_info->total_bytes+=ntohs(ip->ip_len);

	// as a temporary solution, we will issue one flow per one passed packet

	flow *f = pcap_info->fc->match((struct ipv4_key*)&pcap_flow_key, 1, ntohs(ip->ip_len));

/*
	if(FE->Process((u_char *)&pcap_flow_key, 1, ntohs(ip->ip_len), &flow_entry) == -1) {
		IPv4FillFlow((struct ipv4_key*)&pcap_flow_key, flow_entry);
		if(pcap_flow_key.dot1x) {
			struct vlan_info_value *vlan_info = (struct vlan_info_value*)flow_entry->flow->put(ATTR_VLAN_INFO);
			vlan_info->dot1x = pcap_flow_key.dot1x;
		}
#ifdef LAYER7_FILTER
		if (ds->layer7_detect!=LAYER7_DETECT_NONE) layer7_addinfo(pcap_flow_key.tcp_info.dst_port, flow_entry);
#endif
	}
*/
#ifdef URL_CHECK
	if (pcap_info->enable_url_check) {
		int j=layer7_checkinfo(pcap_flow_key.tcp_info.dst_port, f, ip);
		if (j!=0) {
			// printf("URL flow %p: %s\n", f, f->url_value);
			flow *rev_flow = pcap_info->fc->find_rev_flow(f);
			if (rev_flow!=NULL && rev_flow->url_value==NULL) {
				// printf("URL info swap flows %p, %p\n", f, rev_flow);
				rev_flow->url_value=f->url_value;
				rev_flow->url_len = f->url_len;
				f->url_value=NULL;
				f->url_len=0;
			}
		}

	}
#endif

#ifdef ENABLE_DPI
	if (pcap_info->enable_dpi
			#ifdef URL_CHECK
			&& pcap_info->enable_url_check && f->url_len==0
			#endif
		) {
		u_int64_t time = ((uint64_t) hdr->ts.tv_sec) * detection_tick_resolution + hdr->ts.tv_usec / (1000000 / detection_tick_resolution);
		dpi(f, ip, hdr->len - sizeof(struct ethhdr), time);
	}
#endif

}

#endif
