/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	Copyright 1998-2002 Anton Vinokurov <anton@netams.com>
***	Copyright 2002-2008 NeTAMS Development Team
***	This code is GPL v3
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************
**/
/* $Id: layer7.c,v 1.18 2008-02-23 08:35:02 anton Exp $ */

#include "config.h"

#ifdef URL_CHECK

/////////////////////////////////////////////////////////////////////////
int layer7_checkport(u_int16_t port){
		
	switch (htons(port)){
		case 80: 
		case 81: 
		case 8080: 
		case 8000: 
		case 3128: 
			{
			return 1;
			}
	}
	return 0;
}
/////////////////////////////////////////////////////////////////////////
int layer7_checkinfo(u_int16_t port, flow *flow_entry, struct ip *ip){
	int ret=0;
	if (layer7_checkport(port)==0) return 0; // not HTTP traffic
	if (ntohs(ip->ip_len)<50) return 0; // packet too short for URL request
	if (flow_entry->url_value!=NULL) return 0;  // URL info is already set for this flow

	struct tcphdr *th = (struct tcphdr *)((u_char *)ip + ip->ip_hl*4);
	int j=ip->ip_hl*4+th->th_off*4;
	char *http_data = j+(char*)ip;

	char *q1=strcasestr(http_data, "Host:");
	if (q1!=NULL) {
		char *q2=strchr(q1, '\n');
		if (q2!=NULL) {
			char q3[MAX_URL];
			j=q2-q1-6;
			if (j>MAX_URL) j=MAX_URL;
			else if (j<1) return 0;
			snprintf(q3, j, "%s", q1+6);
			flow_entry->url_value=set_string(q3);
			flow_entry->url_len=strlen(q3);
			ret=1;
		}

		q1=strcasestr(http_data, "GET ");
		if (q1!=NULL) {
			q2=strchr(q1+5, ' ');
			if (q2!=NULL) {
				char q3[MAX_URL];
				j=q2-q1-3;
				if (flow_entry->url_value!=NULL) j+=strlen(flow_entry->url_value);
				if (j>MAX_URL) j=MAX_URL;
				else if (j<1) return 0;
				if (strncasecmp(q1, "GET http://", 11)) snprintf(q3, j, "%s%s", flow_entry->url_value?flow_entry->url_value:"", q1+4);
				else snprintf(q3, j>21?j-20:1, "%s", q1+11);
				if (flow_entry->url_value!=NULL) { free(flow_entry->url_value); mem_c_url_free++; mem_url_free+=flow_entry->url_len; }
				char *p=strchr(q3, '\''); for (int i=0; q3[i]!='\0' && p!=NULL; i++, p=strchr(q3+i, '\'')) *p='^';
				p=strchr(q3, '\"'); for (int i=0; q3[i]!='\0' && p!=NULL; i++, p=strchr(q3+i, '\"')) *p='^';
				flow_entry->url_value=set_string(q3);
				flow_entry->url_len=strlen(q3);
				ret=1;
			}
		}
	}
	if (flow_entry->url_len!=0) n4debug("url info: %p ['%s'] %u pkt=%lu bytes=%qu\n", flow_entry, flow_entry->url_value, flow_entry->url_len, flow_entry->packets, flow_entry->bytes);
	return ret;
}
/////////////////////////////////////////////////////////////////////////
#endif /* URL_CHECK */

#ifdef ENABLE_DPI

static struct ipoque_detection_module_struct *ipoque_struct = NULL;
u_int32_t detection_tick_resolution = 1000;
char *prot_short_str[] = { IPOQUE_PROTOCOL_SHORT_STRING };

static void *malloc_wrapper(unsigned long size){	return malloc(size); }
static void free_wrapper(void *freeable){ free(freeable); }

u_int32_t size_id_struct = 0;
u_int32_t size_flow_struct = 0;

void init_dpi(){
	IPOQUE_PROTOCOL_BITMASK all;

	// init global detection structure
	ipoque_struct = ipoque_init_detection_module(detection_tick_resolution, malloc_wrapper, NULL);
	if (ipoque_struct == NULL) {
		n4debug("ERROR: global structure initialization failed\n");
		exit(-1);
	}
	// enable all protocols
	IPOQUE_BITMASK_SET_ALL(all);
	ipoque_set_protocol_detection_bitmask2(ipoque_struct, &all);

	// allocate memory for id and flow tracking
	size_id_struct = ipoque_detection_get_sizeof_ipoque_id_struct();
	size_flow_struct = ipoque_detection_get_sizeof_ipoque_flow_struct();

}

void exit_dpi(){
	ipoque_exit_detection_module(ipoque_struct, free_wrapper);
}

void dpi(flow *flow_entry,  struct ip *ip, int length, u_int64_t time){
	u_int32_t protocol = 0;
	struct ipoque_id_struct *src = NULL;
	struct ipoque_id_struct *dst = NULL;

	n4debug("dpi_enter: flow=%p, len=%d\n", flow_entry, length);
	struct in_addr src_ip, dst_ip;
	memset(&src_ip, 0, sizeof(struct in_addr));
	memset(&dst_ip, 0, sizeof(struct in_addr));
	if (ip->ip_v == 4 && length >= sizeof(struct iphdr)) {
		src_ip.s_addr = ip->ip_src.s_addr;
		dst_ip.s_addr = ip->ip_dst.s_addr;
	} else return;

	//src = get_id(&src_ip);
	//dst = get_id(&dst_ip);

	if (ip->ip_v == 4 && (ip->ip_off & htons(0x1FFF)) != 0) return; // don't count fragments

	protocol = ipoque_detection_process_packet(ipoque_struct, flow_entry->ipoque_flow, (uint8_t *) ip, length, time, src, dst);
	if (protocol!=0) {
		flow_entry->dpi_protocol=protocol;
		n4debug("dpi: flow=%p, proto=%d, name='%s'\n", flow_entry, protocol, prot_short_str[protocol]);
	}
}
#endif /* ENABLE_DPI */
